__version__ = "1.0.2"
__author__ = "Florian Heinrich"
__credits__ = "HEPHY Vienna"

from . import analysis
from . import display
from . import roan_steps
from . import utils
from . import file_io

print(f"APANTIAS version {__version__} loaded.")



import logging
import time
import sys
'''
levels:
logging.debug()
logging.info()
logging.warning()
logging.error()
logging.critical()
'''

class CustomFormatter(logging.Formatter):

    grey = "\x1b[38;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"

    FORMATS = {
        logging.DEBUG: grey + format + reset,
        logging.INFO: grey + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)

class Logger:
    def __init__(self, logger_name: str, 
                 level: str = 'info', file_name: str = None):
        # Create a logger
        self.logger = logging.getLogger(logger_name)
        levels = [logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL]
        if level not in ['debug', 'info', 'warning', 'error', 'critical']:
            raise ValueError('Invalid level')
        level = levels[['debug', 'info', 'warning', 'error', 'critical'].index(level)]
        self.logger.setLevel(level)

        if file_name is not None:
            # Create a file handler
            timestamp = time.strftime('%Y%m%d-%H%M%S')
            fh = logging.FileHandler(timestamp + '_' + file_name)
            fh.setLevel(level)
            # Add the formatter to the handler
            fh.setFormatter(CustomFormatter())
            # Add the handler to the logger
            self.logger.addHandler(fh)
            # print to sys.stdout to avoid red background in notebook
            fh = logging.StreamHandler(sys.stdout)

        # Create a console handler
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(level)
        # Add the formatter to the handler
        ch.setFormatter(CustomFormatter())
        # Add the handler to the logger
        self.logger.addHandler(ch)

    def get_logger(self) -> logging.Logger:
        return self.logger


import numpy as np
from scipy.optimize import curve_fit

from typing import List, Tuple
import gc

from . import logger
from . import fitting
from . import display
from . import utils

_logger = logger.Logger(__name__, "info").get_logger()


def exclude_mips_and_bad_frames(
    data: np.ndarray, thres_mips: int, thres_bad_frames: int
) -> np.ndarray:
    """
    Combines exclude_mips_frames and exclude_bad_frames to exclude frames
    that are above or below the median by a certain threshold.
    MIPS:
    Calculates the median of each frame and deletes frames where there is
    a pixel above or below the median by a certain threshold.
    BAD FRAMES:
    Calculates the average of each frame and excludes frames that are
    above or below the fitted mean by a certain threshold.
    Args:
        data: shape (nframes, column_size, nreps, row_size)
        thres_mips: absolute threshold in adu
        thres_bad_frames: used with the fitted sigma do exclude frames
    Returns:
        np.array in shape (nframes-X, column_size, nreps, row_size)
    """
    # TODO: rethink this. Its quite expensive to calculate the median/mean twice.
    # maybe just keep the more restrictive of the two?
    if np.ndim(data) != 4:
        _logger.error("Input data is not a 4D array.")
        raise ValueError("Input data is not a 4D array.")
    _logger.info(f"Excluding bad frames due to MIPS, threshold: {thres_mips}")
    _logger.info(f"Excluding bad frames, threshold: {thres_bad_frames}")
    median = utils.nanmedian(data, axis=3)
    median = utils.nanmedian(median, axis=2)
    median = utils.nanmedian(median, axis=1)
    # calculate mips mask
    mips_mask = (data > median[:, np.newaxis, np.newaxis, np.newaxis] + thres_mips) | (
        data < median[:, np.newaxis, np.newaxis, np.newaxis] - thres_mips
    )
    mips_mask = np.any(mips_mask, axis=(1, 2, 3))
    _logger.info(f"Excluded {np.sum(mips_mask)} frames due to mips")
    _logger.debug(f"Indices: {np.where(mips_mask)[0]}")
    # calculate bad frames mask
    mean = utils.nanmean(data, axis=3)
    mean = utils.nanmean(mean, axis=2)
    mean = utils.nanmean(mean, axis=1)
    fit = fitting.fit_gauss_to_hist(mean)
    lower_bound = fit[1] - thres_bad_frames * np.abs(fit[2])
    upper_bound = fit[1] + thres_bad_frames * np.abs(fit[2])
    bad_frames_mask = (mean < lower_bound) | (mean > upper_bound)
    _logger.info(f"Excluded {np.sum(bad_frames_mask)} bad frames")
    _logger.debug(f"Indices: {np.where(bad_frames_mask)[0]}")
    mask = mips_mask | bad_frames_mask
    return data[~mask]


def get_slopes(data: np.ndarray) -> np.ndarray:
    """
    Calculates the slope over nreps for every pixel and frame.
    Args:
        data: np.array in shape (nframes, column_size, nreps, row_size)
    Returns:
        slopes: np.array (nframes, column_size, row_size) with the slope values
    """
    if np.ndim(data) != 4:
        _logger.error("Input data is not a 4D array.")
        raise ValueError("Input data is not a 4D array.")
    _logger.info("Calculating slope values")
    slopes = utils.apply_slope_fit_along_frames(data)
    _logger.info("Finished calculating slope values")
    _logger.debug(f"Shape of slopes: {slopes.shape}")
    return slopes


def set_bad_pixellist_to_nan(
    data: np.ndarray, bad_pixels: List[Tuple[int, int]]
) -> None:
    """
    Sets all ignored Pixels in data to NaN. List of pixels is from the
    parameter file. [(row,col), (row,col), ...]
    Args:
        data: np.array in shape (nframes, column_size, nreps, row_size)
        bad_pixels: list of tuples (row,col)
    Returns:
        np.array in shape (nframes, column_size, nreps, row_size)
    """
    if np.ndim(data) != 4:
        _logger.error("Data is not a 4D array")
        raise ValueError("Data is not a 4D array")
    _logger.info("Excluding bad pixels")
    bad_pixel_mask = np.zeros(data.shape, dtype=bool)
    for index in bad_pixels:
        col = index[1]
        row = index[0]
        bad_pixel_mask[:, row, :, col] = True
    data[bad_pixel_mask] = np.nan
    _logger.info(f"Excluded {len(bad_pixels)} pixels")


def correct_common_mode(data: np.ndarray) -> None:
    """
    Calculates the median of euch row in data, and substracts it from
    the row.
    Correction is done inline to save memory.
    Args:
        np.array in shape (nframes, column_size, nreps, row_size)
    """
    if data.ndim != 4:
        _logger.error("Data is not a 4D array")
        raise ValueError("Data is not a 4D array")
    _logger.info("Starting common mode correction.")
    # Iterate over the nframes dimension
    # Calculate the median for one frame
    # median_common = analysis_funcs.parallel_nanmedian_4d_axis3(data)
    # Subtract the median from the frame in-place
    median_common = utils.nanmedian(data, axis=3, keepdims=True)
    data -= median_common
    _logger.info("Data is corrected for common mode.")


# TODO: Rewrite EventMap and GainFit:
# - EventMap: Current structure cannot be written to h5 file
# - GainFit: Find some parameters to find "good" pixels first


def calc_event_map(
    avg_over_nreps: np.ndarray, noise_fitted: np.ndarray, thres_event: int
) -> np.ndarray:
    """
    Finds events in the data by comparing the average over nreps with the
    noise fitted map.
    Args:
        avg_over_nreps: np.array (nframes, column_size, row_size)
        noise_fitted: np.array (column_size, row_size)
        thres_event: threshold for the events
    Returns:
        np.array (column_size, row_size) with lists of signals
    """
    if avg_over_nreps.ndim != 3:
        _logger.error("avg_over_nreps is not a 3D array")
        raise ValueError("avg_over_nreps is not a 3D array")
    if noise_fitted.ndim != 2:
        _logger.error("noise_fitted is not a 2D array")
        raise ValueError("noise_fitted is not a 2D array")
    _logger.info("Finding events")
    threshold_map = noise_fitted * thres_event
    events = avg_over_nreps > threshold_map[np.newaxis, :, :]
    signals = np.full(avg_over_nreps.shape, np.nan)
    signals[events] = avg_over_nreps[events]
    mask = ~np.isnan(signals)
    indices = np.argwhere(mask)
    values = signals[mask]
    _logger.info(f"{indices.shape[0]} events found")
    event_map = np.zeros((64, 64), dtype=object)
    for i in range(event_map.shape[0]):
        for j in range(event_map.shape[1]):
            event_map[i, j] = {"frame": [], "signal": []}
    for index, entry in enumerate(indices):
        frame = int(entry[0])
        row = int(entry[1])
        column = int(entry[2])
        signal = values[index]
        event_map[row][column]["frame"].append(frame)
        event_map[row][column]["signal"].append(signal)
    return event_map


def get_sum_of_event_signals(
    event_map: np.ndarray, row_size: int, column_size: int
) -> np.ndarray:
    """
    Adds the signal values of the events in the event_map
    Args:
        event_map: np.array (column_size, row_size)
        row_size: number of rows in the data
        column_size: number of columns in the data
    Returns:
        np.array (column_size, row_size) with the sum of the signals
    """
    sum_of_events = np.zeros((row_size, column_size))
    for row in range(row_size):
        for column in range(column_size):
            sum_of_events[row][column] = sum(event_map[row][column]["signal"])
    return sum_of_events


def get_sum_of_event_counts(
    event_map: np.ndarray, row_size: int, column_size: int
) -> np.ndarray:
    """
    Sums up the number of events in the event_map
    Args:
        event_map: np.array (column_size, row_size)
        row_size: number of rows in the data
        column_size: number of columns in the data
    Returns:
        np.array (column_size, row_size) with the count of the events
    """
    sum_of_events = np.zeros((row_size, column_size))
    for row in range(row_size):
        for column in range(column_size):
            sum_of_events[row][column] = len(event_map[row][column]["signal"])
    return sum_of_events


def get_gain_fit(
    event_map: np.ndarray, row_size: int, column_size: int, min_signals: int
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Fits a gaussian to the histogram of the signals in the event_map. When
    there are fewer signals than min_signals, or the fit is not
    successfull, the values are set to np.nan.
    Args:
        event_map: np.array (column_size, row_size) with lists of signals
        row_size: number of rows in the data
        column_size: number of columns in the data
        min_signals: minimum number of signals for a fit
    Returns:
        np.array (column_size, row_size) with the mean of the gaussian
        np.array (column_size, row_size) with the sigma of the gaussian
        np.array (column_size, row_size) with the error of the mean
        np.array (column_size, row_size) with the error of the sigma
    """
    # TODO: optimize this
    mean = np.full((row_size, column_size), np.nan)
    sigma = np.full((row_size, column_size), np.nan)
    mean_error = np.full((row_size, column_size), np.nan)
    sigma_error = np.full((row_size, column_size), np.nan)

    def fit_hist(data):

        def gaussian(x, a1, mu1, sigma1):
            return a1 * np.exp(-((x - mu1) ** 2) / (2 * sigma1**2))

        try:
            hist, bins = np.histogram(
                data, bins=10, range=(np.nanmin(data), np.nanmax(data)), density=True
            )
            bin_centers = (bins[:-1] + bins[1:]) / 2
            params, covar = curve_fit(
                gaussian, bin_centers, hist, p0=[1, bins[np.argmax(hist)], 1]
            )
            return (
                params[1],
                abs(params[2]),
                np.sqrt(np.diag(covar))[1],
                np.sqrt(np.diag(covar))[2],
            )
        except:
            return (np.nan, np.nan, np.nan, np.nan)

    count_too_few = 0
    for i in range(event_map.shape[0]):
        for j in range(event_map.shape[1]):
            signals = event_map[i, j]["signal"]
            if len(signals) >= min_signals:
                params = fit_hist(signals)
                mean[i, j] = params[0]
                sigma[i, j] = params[1]
                mean_error[i, j] = params[2]
                sigma_error[i, j] = params[3]
            else:
                count_too_few += 1
    _logger.info(f"{count_too_few} pixels have less than {min_signals} signals")
    return mean, sigma, mean_error, sigma_error



import gc
import os

import numpy as np
from numba import njit, prange

from . import logger
from . import fitting


_logger = logger.Logger(__name__, "info").get_logger()


def get_avg_over_frames(data: np.ndarray) -> np.ndarray:
    """
    Calculates the average over the frames in data.
    Args:
        data: in shape (nframes, column_size, nreps, row_size)

    Returns:
        np.array in shape (column_size, nreps, row_size)
    """
    if np.ndim(data) != 4:
        _logger.error("Input data is not a 4D array.")
        raise ValueError("Input data is not a 4D array.")
    return nanmean(data, axis=0)


def get_avg_over_nreps(data: np.ndarray) -> np.ndarray:
    """
    Calculates the average over the nreps in data.
    Args:
        data: in shape (nframes, column_size, nreps, row_size)
    Returns:
        np.array in shape (nframes, column_size, row_size)
    """
    if np.ndim(data) != 4:
        _logger.error("Input data is not a 4D array.")
        raise ValueError("Input data is not a 4D array.")
    return nanmean(data, axis=2)


def get_avg_over_frames_and_nreps(
    data: np.ndarray,
    avg_over_frames: np.ndarray = None,
    avg_over_nreps: np.ndarray = None,
) -> np.ndarray:
    """
    Calculates the average over the frames and nreps in data. If avg_over_frames
    or avg_over_nreps are already calculated they can be passed as arguments to
    save computation time.
    Args:
        data: in shape (nframes, column_size, nreps, row_size)
        avg_over_frames: (optional) in shape (column_size, nreps, row_size)
        avg_over_nreps: (optional) np.in shape (nframes, column_size, row_size)
    Returns:
        np.array in shape (column_size, row_size)
    """
    if np.ndim(data) != 4:
        _logger.error("Input data is not a 4D array.")
        raise ValueError("Input data is not a 4D array.")

    if avg_over_frames is None and avg_over_nreps is None:
        return parallel_computations.nanmean(
            parallel_computations.nanmean(data, axis=0), axis=2
        )

    if avg_over_frames is not None and avg_over_nreps is not None:
        if np.ndim(avg_over_frames) != 3 or np.ndim(avg_over_nreps) != 3:
            _logger.error("Input avg_over_frames or avg_over_nreps is not a 3D array.")
            raise ValueError(
                "Input avg_over_frames or avg_over_nreps is not a 3D array."
            )
        if avg_over_frames.shape[1] < avg_over_nreps.shape[0]:
            return nanmean(avg_over_frames, axis=1)
        else:
            return nanmean(avg_over_nreps, axis=0)
    else:
        if avg_over_nreps is not None:
            if np.ndim(avg_over_nreps) != 3:
                _logger.error("Input avg_over_nreps is not a 3D array.")
                raise ValueError("Input avg_over_nreps is not a 3D array.")
            return parallel_computations.nanmean(avg_over_nreps, axis=0)
        else:
            if np.ndim(avg_over_frames) != 3:
                _logger.error("Input avg_over_frames is not a 3D array.")
                raise ValueError("Input avg_over_frames is not a 3D array.")
            return parallel_computations.nanmean(avg_over_frames, axis=1)


def get_rolling_average(data: np.ndarray, window_size: int) -> np.ndarray:
    """
    Calculates a rolling average over window_size
    Args:
        data: in 1 dimension
        window_size:
    Returns:
        1D np.array
    """
    weights = np.repeat(1.0, window_size) / window_size
    # Use 'valid' mode to ensure that output has the same length as input
    return np.convolve(data, weights, mode="valid")


def get_ram_usage_in_gb(
    frames: int, column_size: int, nreps: int, row_size: int
) -> int:
    """
    Calculates the RAM usage in GB for a 4D array of the given dimensions.
    Assuming float64. (8 bytes per element)
    """
    return int(frames * column_size * nreps * row_size * 8 / 1024**3) + 1


@njit(parallel=True)
def apply_slope_fit_along_frames(data):
    """
    The equivalent to np.apply_along_axis(func, axis=2, data).
    Args:
        data: 4D np.array
    Returns:
        3D np.array (frame,row,col) with slope value
    """
    if data.ndim != 4:
        _logger.error("Input data is not a 4D array.")
        raise ValueError("Input data is not a 4D array.")
    axis_0_size = data.shape[0]
    axis_1_size = data.shape[1]
    axis_2_size = data.shape[2]
    axis_3_size = data.shape[3]
    output = np.empty((axis_0_size, axis_1_size, axis_3_size))
    for frame in prange(axis_0_size):
        for row in range(axis_1_size):
            for col in range(axis_3_size):
                slope = fitting.linear_fit(data[frame, row, :, col])
                output[frame][row][col] = slope[0]
    return output


def nanmedian(data: np.ndarray, axis: int, keepdims: bool = False) -> np.ndarray:
    """
    The equivalent to np.nanmedian(data, axis=axis, keepdims=keepdims).
    """
    if data.ndim == 2:
        if axis == 0:
            if keepdims:
                return _nanmedian_2d_axis0(data)[np.newaxis, :]
            else:
                return _nanmedian_2d_axis0(data)
        elif axis == 1:
            if keepdims:
                return _nanmedian_2d_axis1(data)[:, np.newaxis]
            else:
                return _nanmedian_2d_axis1(data)
    elif data.ndim == 3:
        if axis == 0:
            if keepdims:
                return _nanmedian_3d_axis0(data)[np.newaxis, :, :]
            else:
                return _nanmedian_3d_axis0(data)
        elif axis == 1:
            if keepdims:
                return _nanmedian_3d_axis1(data)[:, np.newaxis, :]
            else:
                return _nanmedian_3d_axis1(data)
        elif axis == 2:
            if keepdims:
                return _nanmedian_3d_axis2(data)[:, :, np.newaxis]
            else:
                return _nanmedian_3d_axis2(data)
    elif data.ndim == 4:
        if axis == 0:
            if keepdims:
                return _nanmedian_4d_axis0(data)[np.newaxis, :, :, :]
            else:
                return _nanmedian_4d_axis0(data)
        elif axis == 1:
            if keepdims:
                return _nanmedian_4d_axis1(data)[:, np.newaxis, :, :]
            else:
                return _nanmedian_4d_axis1(data)
        elif axis == 2:
            if keepdims:
                return _nanmedian_4d_axis2(data)[:, :, np.newaxis, :]
            else:
                return _nanmedian_4d_axis2(data)
        elif axis == 3:
            if keepdims:
                return _nanmedian_4d_axis3(data)[:, :, :, np.newaxis]
            else:
                return _nanmedian_4d_axis3(data)
    else:
        _logger.error("Data has wrong dimensions")
        return


def nanmean(data: np.ndarray, axis: int, keepdims: bool = False) -> np.ndarray:
    """
    The equivalent to np.nanmean(data, axis=axis, keepdims=keepdims).
    """
    if data.ndim == 2:
        if axis == 0:
            if keepdims:
                return _nanmean_2d_axis0(data)[np.newaxis, :]
            else:
                return _nanmean_2d_axis0(data)
        elif axis == 1:
            if keepdims:
                return _nanmean_2d_axis1(data)[:, np.newaxis]
            else:
                return _nanmean_2d_axis1(data)
    elif data.ndim == 3:
        if axis == 0:
            if keepdims:
                return _nanmean_3d_axis0(data)[np.newaxis, :, :]
            else:
                return _nanmean_3d_axis0(data)
        elif axis == 1:
            if keepdims:
                return _nanmean_3d_axis1(data)[:, np.newaxis, :]
            else:
                return _nanmean_3d_axis1(data)
        elif axis == 2:
            if keepdims:
                return _nanmean_3d_axis2(data)[:, :, np.newaxis]
            else:
                return _nanmean_3d_axis2(data)
    elif data.ndim == 4:
        if axis == 0:
            if keepdims:
                return _nanmean_4d_axis0(data)[np.newaxis, :, :, :]
            else:
                return _nanmean_4d_axis0(data)
        elif axis == 1:
            if keepdims:
                return _nanmean_4d_axis1(data)[:, np.newaxis, :, :]
            else:
                return _nanmean_4d_axis1(data)
        elif axis == 2:
            if keepdims:
                return _nanmean_4d_axis2(data)[:, :, np.newaxis, :]
            else:
                return _nanmean_4d_axis2(data)
        elif axis == 3:
            if keepdims:
                return _nanmean_4d_axis3(data)[:, :, :, np.newaxis]
            else:
                return _nanmean_4d_axis3(data)
    else:
        _logger.error("Data has wrong dimensions")
        return


@njit(parallel=True)
def _nanmedian_4d_axis0(data: np.ndarray) -> np.ndarray:
    """
    The equivalent to np.nanmedian(data, axis=0, keepdims=False).
    Args:
        data: 4D np.array
    Returns:
        3D np.array
    """
    if data.ndim != 4:
        _logger.error("Input data is not a 4D array.")
        raise ValueError("Input data is not a 4D array.")
    axis_1_size = data.shape[1]
    axis_2_size = data.shape[2]
    axis_3_size = data.shape[3]
    output = np.zeros((axis_1_size, axis_2_size, axis_3_size))
    for i in prange(axis_1_size):
        for j in prange(axis_2_size):
            for k in prange(axis_3_size):
                median = np.nanmedian(data[:, i, j, k])
                output[i, j, k] = median
    return output


@njit(parallel=True)
def _nanmedian_4d_axis1(data: np.ndarray) -> np.ndarray:
    """
    The equivalent to np.nanmedian(data, axis=1, keepdims=False).
    Args:
        data: 4D np.array
    Returns:
        3D np.array
    """
    if data.ndim != 4:
        _logger.error("Input data is not a 4D array.")
        raise ValueError("Input data is not a 4D array.")
    axis_0_size = data.shape[0]
    axis_2_size = data.shape[2]
    axis_3_size = data.shape[3]
    output = np.zeros((axis_0_size, axis_2_size, axis_3_size))
    for i in prange(axis_0_size):
        for j in prange(axis_2_size):
            for k in prange(axis_3_size):
                median = np.nanmedian(data[i, :, j, k])
                output[i, j, k] = median
    return output


@njit(parallel=True)
def _nanmedian_4d_axis2(data: np.ndarray) -> np.ndarray:
    """
    The equivalent to np.nanmedian(data, axis=2, keepdims=False).
    Args:
        data: 4D np.array
    Returns:
        3D np.array
    """
    if data.ndim != 4:
        _logger.error("Input data is not a 4D array.")
        raise ValueError("Input data is not a 4D array.")
    axis_0_size = data.shape[0]
    axis_1_size = data.shape[1]
    axis_3_size = data.shape[3]
    output = np.zeros((axis_0_size, axis_1_size, axis_3_size))
    for i in prange(axis_0_size):
        for j in prange(axis_1_size):
            for k in prange(axis_3_size):
                median = np.nanmedian(data[i, j, :, k])
                output[i, j, k] = median
    return output


@njit(parallel=True)
def _nanmedian_4d_axis3(data: np.ndarray) -> np.ndarray:
    """
    The equivalent to np.nanmedian(data, axis=3, keepdims=False).
    Args:
        data: 4D np.array
    Returns:
        3D np.array
    """
    if data.ndim != 4:
        _logger.error("Input data is not a 4D array.")
        raise ValueError("Input data is not a 4D array.")
    axis_0_size = data.shape[0]
    axis_1_size = data.shape[1]
    axis_2_size = data.shape[2]
    output = np.zeros((axis_0_size, axis_1_size, axis_2_size))
    for i in prange(axis_0_size):
        for j in prange(axis_1_size):
            for k in prange(axis_2_size):
                median = np.nanmedian(data[i, j, k, :])
                output[i, j, k] = median
    return output


@njit(parallel=True)
def _nanmedian_3d_axis0(data: np.ndarray) -> np.ndarray:
    """
    The equivalent to np.nanmedian(data, axis=0, keepdims=False).
    Args:
        data: 3D np.array
    Returns:
        2D np.array
    """
    if data.ndim != 3:
        _logger.error("Input data is not a 3D array.")
        raise ValueError("Input data is not a 3D array.")
    axis_1_size = data.shape[1]
    axis_2_size = data.shape[2]
    output = np.zeros((axis_1_size, axis_2_size))
    for i in prange(axis_1_size):
        for j in prange(axis_2_size):
            median = np.nanmedian(data[:, i, j])
            output[i, j] = median
    return output


@njit(parallel=True)
def _nanmedian_3d_axis1(data: np.ndarray) -> np.ndarray:
    """
    The equivalent to np.nanmedian(data, axis=1, keepdims=False).
    Args:
        data: 3D np.array
    Returns:
        2D np.array
    """
    if data.ndim != 3:
        _logger.error("Input data is not a 3D array.")
        raise ValueError("Input data is not a 3D array.")
    axis_0_size = data.shape[0]
    axis_2_size = data.shape[2]
    output = np.zeros((axis_0_size, axis_2_size))
    for i in prange(axis_0_size):
        for j in prange(axis_2_size):
            median = np.nanmedian(data[i, :, j])
            output[i, j] = median
    return output


@njit(parallel=True)
def _nanmedian_3d_axis2(data: np.ndarray) -> np.ndarray:
    """
    The equivalent to np.nanmedian(data, axis=2, keepdims=False).
    Args:
        data: 3D np.array
    Returns:
        2D np.array
    """
    if data.ndim != 3:
        _logger.error("Input data is not a 3D array.")
        raise ValueError("Input data is not a 3D array.")
    axis_0_size = data.shape[0]
    axis_1_size = data.shape[1]
    output = np.zeros((axis_0_size, axis_1_size))
    for i in prange(axis_0_size):
        for j in prange(axis_1_size):
            median = np.nanmedian(data[i, j, :])
            output[i, j] = median
    return output


@njit(parallel=True)
def _nanmedian_2d_axis0(data: np.ndarray) -> np.ndarray:
    """
    The equivalent to np.nanmedian(data, axis=0, keepdims=False).
    Args:
        data: 2D np.array
    Returns:
        1D np.array
    """
    if data.ndim != 2:
        _logger.error("Input data is not a 2D array.")
        raise ValueError("Input data is not a 2D array.")
    axis_1_size = data.shape[1]
    output = np.zeros(axis_1_size)
    for i in prange(axis_1_size):
        median = np.nanmedian(data[:, i])
        output[i] = median
    return output


@njit(parallel=True)
def _nanmedian_2d_axis1(data: np.ndarray) -> np.ndarray:
    """
    The equivalent to np.nanmedian(data, axis=1, keepdims=False).
    Args:
        data: 2D np.array
    Returns:
        1D np.array
    """
    if data.ndim != 2:
        _logger.error("Input data is not a 2D array.")
        raise ValueError("Input data is not a 2D array.")
    axis_0_size = data.shape[0]
    output = np.zeros(axis_0_size)
    for i in prange(axis_0_size):
        median = np.nanmedian(data[i, :])
        output[i] = median
    return output


@njit(parallel=True)
def _nanmean_4d_axis0(data: np.ndarray) -> np.ndarray:
    """
    The equivalent to np.nanmean(data, axis=0, keepdims=False).
    Args:
        data: 4D np.array
    Returns:
        3D np.array
    """
    if data.ndim != 4:
        _logger.error("Input data is not a 4D array.")
        raise ValueError("Input data is not a 4D array.")
    axis_1_size = data.shape[1]
    axis_2_size = data.shape[2]
    axis_3_size = data.shape[3]
    output = np.zeros((axis_1_size, axis_2_size, axis_3_size))
    for i in prange(axis_1_size):
        for j in prange(axis_2_size):
            for k in prange(axis_3_size):
                median = np.nanmean(data[:, i, j, k])
                output[i, j, k] = median
    return output


@njit(parallel=True)
def _nanmean_4d_axis1(data: np.ndarray) -> np.ndarray:
    """
    The equivalent to np.nanmean(data, axis=1, keepdims=False).
    Args:
        data: 4D np.array
    Returns:
        3D np.array
    """
    if data.ndim != 4:
        _logger.error("Input data is not a 4D array.")
        raise ValueError("Input data is not a 4D array.")
    axis_0_size = data.shape[0]
    axis_2_size = data.shape[2]
    axis_3_size = data.shape[3]
    output = np.zeros((axis_0_size, axis_2_size, axis_3_size))
    for i in prange(axis_0_size):
        for j in prange(axis_2_size):
            for k in prange(axis_3_size):
                median = np.nanmean(data[i, :, j, k])
                output[i, j, k] = median
    return output


@njit(parallel=True)
def _nanmean_4d_axis2(data: np.ndarray) -> np.ndarray:
    """
    The equivalent to np.nanmean(data, axis=2, keepdims=False).
    Args:
        data: 4D np.array
    Returns:
        3D np.array
    """
    if data.ndim != 4:
        _logger.error("Input data is not a 4D array.")
        raise ValueError("Input data is not a 4D array.")
    axis_0_size = data.shape[0]
    axis_1_size = data.shape[1]
    axis_3_size = data.shape[3]
    output = np.zeros((axis_0_size, axis_1_size, axis_3_size))
    for i in prange(axis_0_size):
        for j in prange(axis_1_size):
            for k in prange(axis_3_size):
                median = np.nanmean(data[i, j, :, k])
                output[i, j, k] = median
    return output


@njit(parallel=True)
def _nanmean_4d_axis3(data: np.ndarray) -> np.ndarray:
    """
    The equivalent to np.nanmean(data, axis=3, keepdims=False).
    Args:
        data: 4D np.array
    Returns:
        3D np.array
    """
    if data.ndim != 4:
        _logger.error("Input data is not a 4D array.")
        raise ValueError("Input data is not a 4D array.")
    axis_0_size = data.shape[0]
    axis_1_size = data.shape[1]
    axis_2_size = data.shape[2]
    output = np.zeros((axis_0_size, axis_1_size, axis_2_size))
    for i in prange(axis_0_size):
        for j in prange(axis_1_size):
            for k in prange(axis_2_size):
                median = np.nanmean(data[i, j, k, :])
                output[i, j, k] = median
    return output


@njit(parallel=True)
def _nanmean_3d_axis0(data: np.ndarray) -> np.ndarray:
    """
    The equivalent to np.nanmean(data, axis=0, keepdims=False).
    Args:
        data: 3D np.array
    Returns:
        2D np.array
    """
    if data.ndim != 3:
        _logger.error("Input data is not a 3D array.")
        raise ValueError("Input data is not a 3D array.")
    axis_1_size = data.shape[1]
    axis_2_size = data.shape[2]
    output = np.zeros((axis_1_size, axis_2_size))
    for i in prange(axis_1_size):
        for j in prange(axis_2_size):
            median = np.nanmean(data[:, i, j])
            output[i, j] = median
    return output


@njit(parallel=True)
def _nanmean_3d_axis1(data: np.ndarray) -> np.ndarray:
    """
    The equivalent to np.nanmean(data, axis=1, keepdims=False).
    Args:
        data: 3D np.array
    Returns:
        2D np.array
    """
    if data.ndim != 3:
        _logger.error("Input data is not a 3D array.")
        raise ValueError("Input data is not a 3D array.")
    axis_0_size = data.shape[0]
    axis_2_size = data.shape[2]
    output = np.zeros((axis_0_size, axis_2_size))
    for i in prange(axis_0_size):
        for j in prange(axis_2_size):
            median = np.nanmean(data[i, :, j])
            output[i, j] = median
    return output


@njit(parallel=True)
def _nanmean_3d_axis2(data: np.ndarray) -> np.ndarray:
    """
    The equivalent to np.nanmean(data, axis=2, keepdims=False).
    Args:
        data: 3D np.array
    Returns:
        2D np.array
    """
    if data.ndim != 3:
        _logger.error("Input data is not a 3D array.")
        raise ValueError("Input data is not a 3D array.")
    axis_0_size = data.shape[0]
    axis_1_size = data.shape[1]
    output = np.zeros((axis_0_size, axis_1_size))
    for i in prange(axis_0_size):
        for j in prange(axis_1_size):
            median = np.nanmean(data[i, j, :])
            output[i, j] = median
    return output


@njit(parallel=True)
def _nanmean_2d_axis0(data: np.ndarray) -> np.ndarray:
    """
    The equivalent to np.nanmean(data, axis=0, keepdims=False).
    Args:
        data: 2D np.array
    Returns:
        1D np.array
    """
    if data.ndim != 2:
        _logger.error("Input data is not a 2D array.")
        raise ValueError("Input data is not a 2D array.")
    axis_1_size = data.shape[1]
    output = np.zeros(axis_1_size)
    for i in prange(axis_1_size):
        median = np.nanmean(data[:, i])
        output[i] = median
    return output


@njit(parallel=True)
def _nanmean_2d_axis1(data: np.ndarray) -> np.ndarray:
    """
    The equivalent to np.nanmean(data, axis=1, keepdims=False).
    Args:
        data: 2D np.array
    Returns:
        1D np.array
    """
    if data.ndim != 2:
        _logger.error("Input data is not a 2D array.")
        raise ValueError("Input data is not a 2D array.")
    axis_0_size = data.shape[0]
    output = np.zeros(axis_0_size)
    for i in prange(axis_0_size):
        median = np.nanmean(data[i, :])
        output[i] = median
    return output



import os
import fnmatch
import json

from . import logger


class Params:

    _logger = logger.Logger(__name__, "info").get_logger()

    """
    To change/add parameters, edit/add them here.
    Also change the load() function in the RoanSteps class.
    """
    common_params = {
        "common_results_dir": "",  # str
        "common_bad_pixels": [],  # list of tuples
    }
    offnoi_params = {
        "offnoi_data_file": "",  # list of str
        "offnoi_nframes_eval": [0, -1, 1],  # list of ints
        "offnoi_nreps_eval": [0, -1, 1],  # list of ints
        "offnoi_comm_mode": True,  # bool
        "offnoi_thres_mips": 1000,  # float
        "offnoi_thres_bad_frames": 5,  # float
        "offnoi_thres_bad_slopes": 5,  # float
    }
    filter_params = {
        "filter_data_file": "",  # list of str
        "filter_nframes_eval": [0, -1, 1],  # list of ints
        "filter_nreps_eval": [0, -1, 1],  # list of ints
        "filter_comm_mode": True,  # bool
        "filter_thres_mips": 1000,  # float
        "filter_thres_event": 5,  # float
        "filter_use_fitted_offset": False,  # bool
        "filter_thres_bad_frames": 5,  # float
        "filter_thres_bad_slopes": 5,  # float
    }
    gain_params = {"gain_min_signals": 5}  # int

    # types are checked when they are read
    params_types = {
        "common_results_dir": str,
        "common_bad_pixels": list,  # actually, list of tuples
        "offnoi_data_file": str,
        "offnoi_nframes_eval": list,
        "offnoi_nreps_eval": list,
        "offnoi_comm_mode": bool,
        "offnoi_thres_mips": (int, float),
        "offnoi_thres_bad_frames": (int, float),
        "offnoi_thres_bad_slopes": (int, float),
        "filter_data_file": str,
        "filter_nframes_eval": list,
        "filter_nreps_eval": list,
        "filter_comm_mode": bool,
        "filter_thres_mips": (int, float),
        "filter_thres_event": (int, float),
        "filter_use_fitted_offset": bool,
        "filter_thres_bad_frames": (int, float),
        "filter_thres_bad_slopes": (int, float),
        "gain_min_signals": int,
    }

    # required parameters, where there is no default value
    # file cannot be loaded if these are missing
    required_params = [
        "offnoi_bin_file",
        "offnoi_nreps",
        "filter_bin_file",
        "filter_nreps",
    ]

    def __init__(self, json_path: str = None):
        self.default_dict = {
            **self.common_params,
            **self.offnoi_params,
            **self.filter_params,
            **self.gain_params,
        }
        self.inp_dict = None
        self.param_dict = None
        if json_path is not None:
            self.update(json_path)
            self.check_types()
        else:
            self._logger.error("No parameter file provided.")
            self._logger.error("Created default parameter file.")
            self.save_default_file()
            self._logger.error("Add all required parameters to default.")
            self._logger.error("Add the path to the parameter file as an argument.")
            raise ValueError("No parameter file provided.")

    def update(self, json_path: str) -> None:
        try:
            with open(json_path) as f:
                self.inp_dict = json.load(f)
        except:
            self._logger.error("Error loading the parameter file.")
            self.save_default_file()
            self._logger.error(
                "A default parameter file has been saved to the current directory."
            )
            self.param_dict = None
            return
        self.param_dict = self.default_dict.copy()
        # check consistency of the input dict with the default dict
        for key, value in self.inp_dict.items():
            if key not in self.default_dict:
                self._logger.error(f"{key} is not a valid parameter.")
            else:
                self.param_dict[key] = value
        # check for missing parameters, using default if not required
        # if parameter has no default, set param_dict to None
        for key, value in self.param_dict.items():
            if value is None:
                if key in self.required_params:
                    self._logger.error(f"{key} is missing in the file.")
                    self._logger.error("Please provide a complete parameter file")
                    self._logger.error("Run .info() to see the required parameters.")
                    self.param_dict = None
                    break
                else:
                    self._logger.error(
                        f"{key} is missing. Using default: {self.default_dict[key]}"
                    )

    def check_types(self) -> None:
        for key, value in self.param_dict.items():
            if key not in self.params_types:
                raise TypeError(f"There is no type defined for {key}.")
            else:
                expected_type = self.params_types[key]
                if not isinstance(value, expected_type):
                    raise TypeError(f"Expected {key} to be of type {expected_type}.")

    def get_dict(self) -> dict:
        return self.param_dict

    def print_contents(self) -> None:
        for key, value in self.param_dict.items():
            self._logger.info(f"{key}: {value}")

    def info(self) -> None:
        self._logger.info("The following parameters must be provided:")
        self._logger.info("--common parameters:")
        for key in self.common_params.keys():
            self._logger.info(key)
        self._logger.info("--offnoi parameters:")
        for key in self.offnoi_params.keys():
            self._logger.info(key)
        self._logger.info("--filter parameters:")
        for key in self.filter_params.keys():
            self._logger.info(key)
        self._logger.info("--gain parameters:")
        for key in self.gain_params.keys():
            self._logger.info(key)
        self._logger.info("required parameters:")
        for key in self.required_params:
            self._logger.info(key)

    def save_default_file(self, path: str = None) -> None:
        # if no path is provided, save to the current directory
        self._logger.info(f"path: {path}")
        if path is None:
            path = os.path.join(os.getcwd(), "default_params.json")
        else:
            path = os.path.join(path, "default_params.json")
        with open(path, "w") as f:
            json.dump(self.default_dict, f, indent=4)

    def save(self, path: str) -> None:
        with open(path, "w") as f:
            json.dump(self.param_dict, f, indent=4)

    def get_json_file_name_in_folder(self, folder_path: str) -> str:
        count = 0
        json_file = ""
        for file in os.listdir(folder_path):
            if fnmatch.fnmatch(file, "*.json"):
                json_file = os.path.join(folder_path, file)
                count += 1
        if count == 1:
            return json_file
        else:
            return None



import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np

from . import logger
from . import fitting

_logger = logger.Logger(__name__, "info").get_logger()


def draw_hist(
    data: np.ndarray,
    file_name: str = "histogram",
    save_to: str = None,
    log: bool = False,
    **kwargs,
) -> None:
    """
    Draw a histogram of the data. If a folder is provided, the plot is saved.
    Args:
        data: np.array in any shape
        file_name: str (optional) default is "histogram"
        save_to: str (optional) default is None
        log: bool (optional) default
        **kwargs: (optional) passed to plt.hist
    """
    plt.clf()
    plt.hist(data.ravel(), **kwargs)
    if log:
        plt.yscale("log")
    if save_to is not None:
        plt.savefig(save_to + file_name + ".png")
        plt.close()
    else:
        plt.show()


def draw_heatmap(
    data: np.ndarray, file_name: str = "heatmap", save_to: str = None, **kwargs
) -> None:
    """
    Draw a heatmap of the data. If a folder is provided, the plot is saved.
    Args:
        data: np.array in 2 dimensions
        file_name
        save_to
        **kwargs: is passed to plt.hist
    """
    if data.ndim != 2:
        _logger.error("Input data is not a 2D array.")
        raise ValueError("Input data is not a 2D array.")
    plt.clf()
    cmap = kwargs.get("cmap", "coolwarm")
    sns.heatmap(data, cmap=cmap, **kwargs)
    if save_to is not None:
        plt.savefig(save_to + file_name + ".png")
        plt.close()
    else:
        plt.show()


def draw_graph(
    data: np.ndarray, file_name: str = "graph", save_to: str = None, **kwargs
) -> None:
    """
    Draw a graph of the data. If a folder is provided, the plot is saved.
    """
    plt.clf()
    plt.plot(data.ravel(), **kwargs)
    if save_to is not None:
        plt.savefig(save_to + file_name + ".png")
        plt.close()
    else:
        plt.show()


def draw_hist_and_gauss_fit(
    data: np.ndarray,
    bins: str,
    amplitude: float,
    mean: float,
    sigma: float,
    file_name: str = None,
    save_to: str = None,
) -> None:
    """
    Draw a histogram of the data and a gaussian fit
    """
    plt.clf()
    hist, hist_bins = np.histogram(
        data, bins, range=(np.nanmin(data), np.nanmax(data)), density=True
    )
    bin_centers = (hist_bins[:-1] + hist_bins[1:]) / 2
    plt.hist(data, bins=hist_bins, density=True, alpha=0.5)
    plt.plot(bin_centers, fitting.gaussian(bin_centers, amplitude, mean, sigma), "r-")
    plt.title(f"Fitted parameters:\nMean: {mean:.2f}\nSigma: {sigma:.2f}")
    plt.xlabel("Value")
    plt.ylabel("Frequency")
    if save_to is not None:
        if file_name is None:
            file_name = "hist_and_fit"
        plt.savefig(save_to + "/" + file_name + ".png")
        plt.close()
    else:
        plt.show()



import gc
import os
import psutil
from datetime import datetime

import numpy as np

from . import logger
from . import utils
from . import analysis as an
from . import params
from . import fitting as fit
from . import file_io as io


class RoanSteps:
    _logger = logger.Logger("nproan-RoanSteps", "info").get_logger()

    def __init__(self, prm_file: str, ram: int) -> None:
        self.ram_available = ram
        self.load(prm_file)

    def load(self, prm_file: str) -> None:
        # load parameter file
        self.params = params.Params(prm_file)
        self.params_dict = self.params.get_dict()

        # polarity is from the old code, im not quite sure why it is -1
        self.polarity = -1

        # common parameters from params file
        self.results_dir = self.params_dict["common_results_dir"]
        self.bad_pixels = self.params_dict["common_bad_pixels"]

        # offnoi parameters from params file
        self.offnoi_data_file = self.params_dict["offnoi_data_file"]
        self.offnoi_nframes_eval = self.params_dict["offnoi_nframes_eval"]
        self.offnoi_nreps_eval = self.params_dict["offnoi_nreps_eval"]
        self.offnoi_comm_mode = self.params_dict["offnoi_comm_mode"]
        self.offnoi_thres_mips = self.params_dict["offnoi_thres_mips"]
        self.offnoi_thres_bad_frames = self.params_dict["offnoi_thres_bad_frames"]
        self.offnoi_thres_bad_slopes = self.params_dict["offnoi_thres_bad_slopes"]

        # filter parameters from params file
        self.filter_data_file = self.params_dict["filter_data_file"]
        self.filter_nframes_eval = self.params_dict["filter_nframes_eval"]
        self.filter_nreps_eval = self.params_dict["filter_nreps_eval"]
        self.filter_comm_mode = self.params_dict["filter_comm_mode"]
        self.filter_thres_mips = self.params_dict["filter_thres_mips"]
        self.filter_thres_event = self.params_dict["filter_thres_event"]
        self.filter_use_fitted_offset = self.params_dict["filter_use_fitted_offset"]
        self.filter_thres_bad_frames = self.params_dict["filter_thres_bad_frames"]
        self.filter_thres_bad_slopes = self.params_dict["filter_thres_bad_slopes"]

        # parameters from data_h5 files
        total_frames_offnoi, column_size_offnoi, row_size_offnoi, nreps_offnoi = (
            io.get_params_from_data_file(self.offnoi_data_file)
        )
        total_frames_filter, column_size_filter, row_size_filter, nreps_filter = (
            io.get_params_from_data_file(self.filter_data_file)
        )
        # check if sensor size is equal
        if (
            column_size_offnoi != column_size_filter
            or row_size_offnoi != row_size_filter
        ):
            raise ValueError(
                "Column size or row size of offnoi and filter data files are not equal."
            )

        self.column_size = column_size_offnoi
        self.row_size = row_size_offnoi
        # set total number of frames and nreps from the data file
        self.offnoi_total_nreps = nreps_offnoi
        self.offnoi_total_frames = total_frames_offnoi
        self.filter_total_nreps = nreps_filter
        self.filter_total_frames = total_frames_filter

        # nreps_eval and nframes_eval is [start,stop,step], if stop is -1 it goes to the end
        if self.offnoi_nframes_eval[1] == -1:
            self.offnoi_nframes_eval[1] = self.offnoi_total_frames
        if self.offnoi_nreps_eval[1] == -1:
            self.offnoi_nreps_eval[1] = self.offnoi_total_nreps
        if self.filter_nframes_eval[1] == -1:
            self.filter_nframes_eval[1] = self.filter_total_frames
        if self.filter_nreps_eval[1] == -1:
            self.filter_nreps_eval[1] = self.filter_total_nreps

        # create slices, these are used to get the data from the data files
        self.offnoi_nreps_slice = slice(*self.offnoi_nreps_eval)
        self.offnoi_nframes_slice = slice(*self.offnoi_nframes_eval)
        self.filter_nreps_slice = slice(*self.filter_nreps_eval)
        self.filter_nframes_slice = slice(*self.filter_nframes_eval)

        # set variables to number of nreps_eval and nframes_eval to be evaluated (int)
        self.offnoi_nreps_eval = int(
            (self.offnoi_nreps_eval[1] - self.offnoi_nreps_eval[0])
            / self.offnoi_nreps_eval[2]
        )
        self.offnoi_nframes_eval = int(
            (self.offnoi_nframes_eval[1] - self.offnoi_nframes_eval[0])
            / self.offnoi_nframes_eval[2]
        )
        self.filter_nreps_eval = int(
            (self.filter_nreps_eval[1] - self.filter_nreps_eval[0])
            / self.filter_nreps_eval[2]
        )
        self.filter_nframes_eval = int(
            (self.filter_nframes_eval[1] - self.filter_nframes_eval[0])
            / self.filter_nframes_eval[2]
        )

        # check, if offnoi_nreps_eval is greater or equal than filter_nreps_eval
        # this is necessary, because the filter step needs the offset_raw from the offnoi step
        if self.offnoi_nreps_eval < self.filter_nreps_eval:
            raise ValueError(
                "offnoi_nreps_eval must be greater or equal than filter_nreps_eval"
            )

        # create analysis h5 file
        timestamp = datetime.now().strftime("%Y%m%d-%H%M%S")
        bin_filename = os.path.basename(self.offnoi_data_file)[:-3]
        self.analysis_file_name = f"{timestamp}_{bin_filename}.h5"
        self.analysis_file = os.path.join(self.results_dir, self.analysis_file_name)
        io.create_analysis_file(
            self.results_dir,
            self.analysis_file_name,
            self.offnoi_data_file,
            self.filter_data_file,
            self.params_dict,
        )
        self._logger.info(
            f"Created analysis h5 file: {self.results_dir}/{self.analysis_file_name}"
        )
        self._logger.info(f"Parameters loaded:")
        self.params.print_contents()

    def calc_offnoi_step(self) -> None:

        estimated_ram_usage = (
            utils.get_ram_usage_in_gb(
                self.offnoi_nframes_eval,
                self.column_size,
                self.offnoi_nreps_eval,
                self.row_size,
            )
            * 2.5  # this is estimated, better safe than sorry
        )
        self._logger.info(f"RAM available: {self.ram_available:.1f} GB")
        self._logger.info(f"Estimated RAM usage: {estimated_ram_usage:.1f} GB")
        steps_needed = int(estimated_ram_usage / self.ram_available) + 1
        self._logger.info(f"Steps needed: {steps_needed}")

        # (planned) frames per step, so that ram usage is below the available ram
        frames_per_step = int(self.offnoi_nframes_eval / steps_needed)

        # total processed frames over all steps
        total_frames_processed = 0

        for step in range(steps_needed):
            self._logger.info(f"Start processing step {step+1} of {steps_needed}")
            current_frame_slice = slice(
                total_frames_processed,
                total_frames_processed + frames_per_step,
            )
            slices = [
                current_frame_slice,
                slice(None),
                self.offnoi_nreps_slice,
                slice(None),
            ]
            data = (
                io.get_data_from_file(self.offnoi_data_file, "", "data", slices)
                * self.polarity
            )
            self._logger.info(f"Data loaded: {data.shape}")
            # set values of all frames and nreps of bad pixels to nan
            if self.bad_pixels:
                data = an.set_bad_pixellist_to_nan(data, self.bad_pixels)
            # delete bad frames from data
            if self.offnoi_thres_bad_frames != 0 or self.offnoi_thres_mips != 0:
                data = an.exclude_mips_and_bad_frames(
                    data, self.offnoi_thres_mips, self.offnoi_thres_bad_frames
                )
                self._logger.debug(f"Shape of data: {data.shape}")
            # Calculate offset_raw on the raw data and update file
            avg_over_frames = utils.get_avg_over_frames(data)
            io.add_array(self.analysis_file, "offnoi", "offset_raw", avg_over_frames)
            # offset the data and correct for common mode if necessary
            data -= avg_over_frames[np.newaxis, :, :, :]
            if self.offnoi_comm_mode is True:
                an.correct_common_mode(data)
            # calculate rndr signals and update file
            avg_over_nreps = utils.get_avg_over_nreps(data)
            io.add_array(self.analysis_file, "offnoi", "rndr_signals", avg_over_nreps)
            # calculate bad slopes and update file
            if self.offnoi_thres_bad_slopes != 0:
                slopes = an.get_slopes(data)
                io.add_array(self.analysis_file, "offnoi", "slopes", slopes)
            total_frames_processed += frames_per_step
            self._logger.info(f"Finished step {step+1} of {steps_needed} total Steps")

        # TODO: paralellize this
        slopes = io.get_data_from_file(self.analysis_file, "offnoi", "slopes")
        bad_slopes_pos = np.full(slopes.shape, False, dtype=bool)
        fit_params_list = []

        for row in range(slopes.shape[1]):
            for col in range(slopes.shape[2]):
                slopes_pixelwise = slopes[:, row, col]
                fit_pixelwise = fit.fit_gauss_to_hist(slopes_pixelwise.flatten())
                lower_bound = fit_pixelwise[1] - self.offnoi_thres_bad_slopes * np.abs(
                    fit_pixelwise[2]
                )
                upper_bound = fit_pixelwise[1] + self.offnoi_thres_bad_slopes * np.abs(
                    fit_pixelwise[2]
                )
                bad_slopes_mask = (slopes_pixelwise < lower_bound) | (
                    slopes_pixelwise > upper_bound
                )
                frame = np.where(bad_slopes_mask)[0]
                bad_slopes_pos[frame, row, col] = True
                fit_params_list.append(fit_pixelwise)
        bad_slopes_fit = np.vstack(fit_params_list)

        io.add_array(self.analysis_file, "offnoi", "bad_slopes_pos", bad_slopes_pos)
        io.add_array(self.analysis_file, "offnoi", "bad_slopes_fit", bad_slopes_fit)

        avg_over_nreps = io.get_data_from_file(
            self.analysis_file, "offnoi", "rndr_signals"
        )
        avg_over_nreps_slopes_removed = avg_over_nreps.copy()
        avg_over_nreps_slopes_removed[bad_slopes_pos] = np.nan
        io.add_array(
            self.analysis_file,
            "offnoi",
            "rndr_signals_slopes_removed",
            avg_over_nreps_slopes_removed,
        )

        self._logger.info("Fitting pixelwise for offset and noise")
        fitted = fit.get_fit_gauss(avg_over_nreps)
        io.add_array(self.analysis_file, "offnoi", "fit", fitted)
        fitted = fit.get_fit_gauss(avg_over_nreps_slopes_removed)
        io.add_array(self.analysis_file, "offnoi", "fit_slopes_removed", fitted)

    # TODO: continue here
    def calc_filter_step(self) -> None:
        estimated_ram_usage = (
            utils.get_ram_usage_in_gb(
                self.filter_nframes_eval,
                self.column_size,
                self.filter_nreps_eval,
                self.row_size,
            )
            * 2.5  # this is estimated, better safe than sorry
        )
        self._logger.info(f"RAM available: {self.ram_available:.1f} GB")
        self._logger.info(f"Estimated RAM usage: {estimated_ram_usage:.1f} GB")
        steps_needed = int(estimated_ram_usage / self.ram_available) + 1
        self._logger.info(f"Steps needed: {steps_needed}")

        # (planned) frames per step, so that ram usage is below the available ram
        frames_per_step = int(self.filter_nframes_eval / steps_needed)

        # total processed frames over all steps
        total_frames_processed = 0

        for step in range(steps_needed):
            self._logger.info(f"Start processing step {step+1} of {steps_needed}")
            current_frame_slice = slice(
                total_frames_processed,
                total_frames_processed + frames_per_step,
            )
            slices = [
                current_frame_slice,
                slice(None),
                self.filter_nreps_slice,
                slice(None),
            ]
            data = (
                io.get_data_from_file(self.filter_data_file, "", "data", slices)
                * self.polarity
            )
            self._logger.info(f"Data loaded: {data.shape}")
            # set values of all frames and nreps of bad pixels to nan
            if self.bad_pixels:
                data = an.set_bad_pixellist_to_nan(data, self.bad_pixels)
            # delete bad frames from data
            if self.filter_thres_bad_frames != 0 or self.filter_thres_mips != 0:
                data = an.exclude_mips_and_bad_frames(
                    data, self.filter_thres_mips, self.filter_thres_bad_frames
                )
            self._logger.debug(f"Shape of data: {data.shape}")

            self.final_frames_per_step = data.shape[0]
            # Get offset_raw from offnoi step
            avg_over_frames_offnoi = io.get_data_from_file(
                self.analysis_file, "offnoi", "offset_raw"
            )
            # offset the data and correct for common mode if necessary
            data -= avg_over_frames_offnoi[np.newaxis, :, : self.filter_nreps_eval, :]
            if self.filter_comm_mode is True:
                an.correct_common_mode(data)
            # calculate rndr signals and update file
            avg_over_nreps = utils.get_avg_over_nreps(data)
            # subtract fitted offset from data
            fitted_offset = io.get_data_from_file(self.analysis_file, "offnoi", "fit")[
                1
            ]
            avg_over_nreps -= fitted_offset
            io.add_array(self.analysis_file, "filter", "rndr_signals", avg_over_nreps)
            # calculate bad slopes and update file
            if self.filter_thres_bad_slopes != 0:
                slopes = an.get_slopes(data)
                io.add_array(self.analysis_file, "filter", "slopes", slopes)
            self._logger.info(f"Finished step {step+1} of {steps_needed} total Steps")
            total_frames_processed += frames_per_step

        slopes = io.get_data_from_file(self.analysis_file, "filter", "slopes")
        bad_slopes_pos = np.full(slopes.shape, False, dtype=bool)
        fit_params_list = []

        for row in range(slopes.shape[1]):
            for col in range(slopes.shape[2]):
                slopes_pixelwise = slopes[:, row, col]
                fit_pixelwise = fit.fit_gauss_to_hist(slopes_pixelwise.flatten())
                lower_bound = fit_pixelwise[1] - self.offnoi_thres_bad_slopes * np.abs(
                    fit_pixelwise[2]
                )
                upper_bound = fit_pixelwise[1] + self.offnoi_thres_bad_slopes * np.abs(
                    fit_pixelwise[2]
                )
                bad_slopes_mask = (slopes_pixelwise < lower_bound) | (
                    slopes_pixelwise > upper_bound
                )
                frame = np.where(bad_slopes_mask)[0]
                bad_slopes_pos[frame, row, col] = True
                fit_params_list.append(fit_pixelwise)
        bad_slopes_fit = np.vstack(fit_params_list)
        io.add_array(self.analysis_file, "filter", "bad_slopes_pos", bad_slopes_pos)
        io.add_array(self.analysis_file, "filter", "bad_slopes_fit", bad_slopes_fit)

        avg_over_nreps_final = io.get_data_from_file(
            self.analysis_file, "filter", "rndr_signals"
        )
        avg_over_nreps_slopes_removed = avg_over_nreps_final.copy()
        avg_over_nreps_slopes_removed[bad_slopes_pos] = np.nan
        io.add_array(
            self.analysis_file,
            "filter",
            "rndr_signals_slopes_removed",
            avg_over_nreps_slopes_removed,
        )



import gc

from scipy.optimize import curve_fit
import numpy as np
from numba import njit, prange

from . import logger

_logger = logger.Logger(__name__, "info").get_logger()


# TODO: Optimize this by not computing the guesses (calculate them beforehand und save them to a file)
def fit_gauss_to_hist(data_to_fit: np.ndarray) -> np.ndarray:
    """
    fits a gaussian to a histogram using the scipy curve_fit method

    Args:
        data_to_fit: np.array in 1 dimension
    Returns:
        np.array[amplitude, mean, sigma, error_amplitude, error_mean, error_sigma]
    """
    guess = [1, np.nanmedian(data_to_fit), np.nanstd(data_to_fit)]
    try:
        hist, bins = np.histogram(
            data_to_fit,
            bins=100,
            range=(np.nanmin(data_to_fit), np.nanmax(data_to_fit)),
            density=True,
        )
        bin_centers = (bins[:-1] + bins[1:]) / 2
        params, covar = curve_fit(gaussian, bin_centers, hist, p0=guess)
        return np.array(
            [
                params[0],
                params[1],
                np.abs(params[2]),
                np.sqrt(np.diag(covar))[0],
                np.sqrt(np.diag(covar))[1],
                np.sqrt(np.diag(covar))[2],
            ]
        )
    except:
        _logger.debug("Fitting for this histogram failed. Returning NaNs.")
        return np.array([np.nan, np.nan, np.nan, np.nan, np.nan, np.nan])


def get_fit_gauss(data: np.ndarray) -> np.ndarray:
    """
    fits a gaussian for every pixel. The fitting is done over the
    histogram of the pixel values from all frames using the scipy
    curve_fit method.

    Args:
        data: in shape (nframes, column_size, row_size)
    Returns:
        np.array in shape (6, rows, columns)
        index 0: amplitude
        index 1: mean
        index 2: sigma
        index 3: error_amplitude
        index 4: error_mean
        index 5: error_sigma
    """
    if data.ndim != 3:
        _logger.error("Data is not a 3D array")
        raise ValueError("Data is not a 3D array")
    # apply the function to every frame
    output = np.apply_along_axis(fit_gauss_to_hist, axis=0, arr=data)
    return output


def gaussian(x: float, a1: float, mu1: float, sigma1: float) -> float:
    return a1 * np.exp(-((x - mu1) ** 2) / (2 * sigma1**2))


def two_gaussians(
    x: float, a1: float, mu1: float, sigma1: float, a2: float, mu2: float, sigma2: float
) -> float:
    return a1 * np.exp(
        -((x - mu1) ** 2) / (2 * sigma1**2)
        + a2 * np.exp(-((x - mu2) ** 2) / (2 * sigma2**2))
    )


@njit(parallel=False)
def linear_fit(data: np.ndarray) -> np.ndarray:
    """
    Fits a linear function to the data using the least squares method.
    """
    x = np.arange(data.size)
    n = data.size

    # Calculate the sums needed for the linear fit
    sum_x = np.sum(x)
    sum_y = np.sum(data)
    sum_xx = np.sum(x * x)
    sum_xy = np.sum(x * data)

    # Calculate the slope (k) and intercept (d)
    k = (n * sum_xy - sum_x * sum_y) / (n * sum_xx - sum_x * sum_x)
    d = (sum_y - k * sum_x) / n

    return np.array([k, d])



import h5py
import numpy as np
import json
from typing import List, Tuple, Optional
import os

from . import logger

_logger = logger.Logger(__name__, "info").get_logger()


def read_data_chunk_from_bin(
    bin_file: str,
    column_size: int,
    row_size: int,
    key_ints: int,
    nreps: int,
    frames_to_read: int,
    offset: int,
) -> Tuple[np.ndarray, int]:
    """
    Reads data from a .bin file in chunks and returns a numpy array with the data and the new offset.
    The bins usually have a header of 8 bytes, followed by the data. So an initial offset of 8 is needed.
    After that the new offset is the position in the file where the next chunk should start. The parameter
    frames_to_read should be set to a value that fits into the available RAM.
    The function returns -1 as offset when end of file is reached.

    Returns:
        data, offset: Tuple[np.ndarray, int]
    """
    # test if nreps make sense
    if offset == 8:
        test_data = np.fromfile(
            bin_file,
            dtype="uint16",
            count=column_size * raw_row_size * 2000 * 2 * 10,
            offset=offset,
        )
        test_data = test_data.reshape(-1, raw_row_size)
        frame_keys = np.where(test_data[:, column_size] == 65535)
        frames = np.stack((frame_keys[0][:-1], frame_keys[0][1:]))
        diff = np.diff(frames, axis=0)
        median_diff = np.median(diff)
        estimated_nreps = int(median_diff / column_size)
        if nreps != estimated_nreps:
            _logger.warning(f"Estimated nreps: {estimated_nreps}, given nreps: {nreps}")
            _logger.warning("Proceeding using estimated nreps")
            nreps = estimated_nreps

    raw_row_size = row_size + key_ints
    raw_frame_size = column_size * raw_row_size * nreps
    rows_per_frame = column_size * nreps
    chunk_size = raw_frame_size * frames_to_read

    inp_data = np.fromfile(bin_file, dtype="uint16", count=chunk_size, offset=offset)
    offset += chunk_size * 2  # offset is in bytes, uint16 = 16 bit = 2 bytes
    # check if file is at its end
    if inp_data.size == 0:
        return None, -1
    # reshape the array into rows -> (#ofRows,67)
    inp_data = inp_data.reshape(-1, raw_row_size)
    # find all the framekeys
    frame_keys = np.where(inp_data[:, column_size] == 65535)
    # stack them and calculate difference to find incomplete frames
    frames = np.stack((frame_keys[0][:-1], frame_keys[0][1:]))
    diff = np.diff(frames, axis=0)
    valid_frames_position = np.nonzero(diff == rows_per_frame)[1]
    if len(valid_frames_position) == 0:
        _logger.error("No valid frames found in chunk, wrong nreps?")
        raise Exception("No valid frames found in chunk, wrong nreps?")
    valid_frames = frames.T[valid_frames_position]
    frame_start_indices = valid_frames[:, 0]
    frame_end_indices = valid_frames[:, 1]
    inp_data = np.array(
        [
            inp_data[start + 1 : end + 1, :64]
            for start, end in zip(frame_start_indices, frame_end_indices)
        ]
    )
    inp_data = inp_data.reshape(-1, column_size, nreps, row_size)
    return inp_data, offset


def create_data_file_from_bins(
    bin_files: List[str],
    output_folder: str,
    output_filename: str,
    nreps: int,
    column_size: int = 64,
    row_size: int = 64,
    attributes: dict = None,
    compression: str = None,
    available_ram_gb: int = 16,
) -> None:
    """
    This function creates a data file in h5 format from a list of bin files.
    It reads chunks from the bin files to memory and writes them to the h5 file.
    The chunk size is determined by the available RAM in GB. A chunk of 30% of
    the available RAM is read at once.
    Args:
        bin_files: absolute paths to the bin files
        output_folder: absolute path to a output folder
        output_filename
        nreps
        column_size
        row_size
        attributes
        compression
        available_ram_gb: the available RAM in GB
    """
    output_file = os.path.join(output_folder, output_filename)
    # check if h5 file already exists
    if os.path.exists(output_file):
        _logger.error(f"File {output_file} already exists. Please delete")
        raise Exception(f"File {output_file} already exists. Please delete")
    # check if bin files exist
    for bin_file in bin_files:
        if not os.path.exists(bin_file):
            _logger.error(f"File {bin_file} does not exist")
            raise Exception(f"File {bin_file} does not exist")
    # create the hdf5 file
    with h5py.File(output_file, "w") as f:
        # create the dataset
        dataset = f.create_dataset(
            "data",
            dtype="uint16",
            shape=(0, column_size, nreps, row_size),
            maxshape=(None, column_size, nreps, row_size),
            chunks=(1, column_size, nreps, row_size),
            compression=compression,
        )
        f.attrs["description"] = (
            "This file contains the raw data from the bin files, only complete frames are saved"
        )
        dataset.attrs["bin_files"] = bin_files
        dataset.attrs["column_size"] = column_size
        dataset.attrs["row_size"] = row_size
        dataset.attrs["nreps"] = nreps
        dataset.attrs["total_frames"] = 0
        if compression:
            dataset.attrs["compression"] = compression
        else:
            dataset.attrs["compression"] = "None"
        if attributes:
            for key, value in attributes.items():
                dataset.attrs[key] = value
        _logger.info(f"Initialized empty file: {output_file}")

        for bin_file in bin_files:
            file_size = os.path.getsize(bin_file)
            frame_size_bytes = column_size * row_size * nreps * 2
            file_size_gb = file_size / (1024 * 1024 * 1024)
            estimated_frames = file_size / frame_size_bytes
            # determine how many frames to read at once
            frames_to_read = int(
                (available_ram_gb * 1024 * 1024 * 1024 / frame_size_bytes) * 0.3
            )
            chunk_size = (frames_to_read * frame_size_bytes) / (1024 * 1024 * 1024)
            _logger.info(
                f"Loading file: {bin_file}\n size: {file_size_gb:.2f} GB\n estimated frames: {estimated_frames:.1f}\n chunk size: {chunk_size:.2f} GB"
            )
            offset = 8
            while offset != -1:  # get_data returns -1 as offset when EOF is reached
                try:
                    new_data, new_offset = read_data_chunk_from_bin(
                        bin_file,
                        column_size,
                        row_size,
                        3,
                        nreps,
                        frames_to_read,
                        offset,
                    )
                except Exception as e:
                    _logger.error(e)
                    _logger.error(f"Deleting file: {output_file}")
                    # delete the h5 file
                    os.remove(output_file)
                    break
                offset = new_offset
                if new_data is not None:
                    dataset.resize(dataset.shape[0] + new_data.shape[0], axis=0)
                    # Append the new data
                    dataset[-new_data.shape[0] :] = new_data
                    frames_loaded = dataset.shape[0]
                    _logger.info(
                        f"progress: {frames_loaded}/{estimated_frames:.0f} frames loaded ({frames_loaded/estimated_frames:.2%})"
                    )
        dataset.attrs["total_frames"] = dataset.shape[0]


def display_file_structure(file_path: str) -> None:
    """
    Displays the structure (groups and datasets) of an HDF5 file.
    """

    def print_structure(name, obj):
        indent = "  " * (name.count("/") - 1)
        if isinstance(obj, h5py.Group):
            print(f"{indent}Group: {name}")
        elif isinstance(obj, h5py.Dataset):
            print(f"{indent}Dataset: {name}, shape: {obj.shape}, dtype: {obj.dtype}")

        # Print attributes
        for key, value in obj.attrs.items():
            print(f"{indent}  Attribute: {key} = {value}")

    with h5py.File(file_path, "r") as file:
        file.visititems(print_structure)


def get_params_from_data_file(file_path: str) -> Tuple[int, int, int, int]:
    """
    Get the parameters from the data h5 file.
    """
    with h5py.File(file_path, "r") as file:
        total_frames = file["data"].shape[0]
        column_size = file["data"].attrs["column_size"]
        row_size = file["data"].attrs["row_size"]
        nreps = file["data"].attrs["nreps"]
    return total_frames, column_size, row_size, nreps


def create_analysis_file(
    output_folder: str,
    output_filename: str,
    offnoi_data_file: str,
    filter_data_file: str,
    parameter_file_contents: dict,
) -> None:
    """
    Create an analysis h5 file with offnoi/filter/gain groups.
    An existing data file must be provided for offnoi and filter.
    The parameter file contents are saved as a json string.
    """

    output_file = os.path.join(output_folder, output_filename)
    if os.path.exists(output_file):
        raise Exception(f"File {output_file} already exists. Please delete")
    # raise excption when data files dont exist
    if os.path.exists(offnoi_data_file) == False:
        _logger.error(f"File {offnoi_data_file} does not exist.")
        raise Exception(f"File {offnoi_data_file} does not exist.")
    if os.path.exists(filter_data_file) == False:
        _logger.error(f"File {filter_data_file} does not exist.")
        raise Exception(f"File {filter_data_file} does not exist.")
    # create the hdf5 file
    with h5py.File(output_file, "w") as f:
        f.attrs["description"] = "This file contains the results of the analysis."

        f.create_group("offnoi")
        f["offnoi"].attrs["data"] = offnoi_data_file
        f["offnoi"].attrs[
            "description"
        ] = "This group contains the results of the offset noise analysis."
        with h5py.File(offnoi_data_file, "r") as offnoi_data_file:
            f["offnoi"].attrs["bin_files"] = offnoi_data_file["data"].attrs["bin_files"]
            f["offnoi"].attrs["column_size"] = offnoi_data_file["data"].attrs[
                "column_size"
            ]
            f["offnoi"].attrs["row_size"] = offnoi_data_file["data"].attrs["row_size"]
            f["offnoi"].attrs["nreps"] = offnoi_data_file["data"].attrs["nreps"]
            f["offnoi"].attrs["total_frames"] = offnoi_data_file["data"].shape[0]

        f.create_group("filter")
        f["filter"].attrs["data"] = filter_data_file
        f["filter"].attrs[
            "description"
        ] = "This group contains the results of the filter analysis."
        with h5py.File(filter_data_file, "r") as filter_data_file:
            f["filter"].attrs["bin_files"] = filter_data_file["data"].attrs["bin_files"]
            f["filter"].attrs["column_size"] = filter_data_file["data"].attrs[
                "column_size"
            ]
            f["filter"].attrs["row_size"] = filter_data_file["data"].attrs["row_size"]
            f["filter"].attrs["nreps"] = filter_data_file["data"].attrs["nreps"]
            f["filter"].attrs["total_frames"] = filter_data_file["data"].shape[0]

        f.create_group("gain")
        f["gain"].attrs[
            "description"
        ] = "This group contains the results of the gain analysis."
        f.create_group("parameter_json")

        f["parameter_json"].attrs[
            "description"
        ] = "This group contains the parameter file used for the analysis."
        f.create_dataset(
            "parameter_json/parameter_file",
            data=repr(parameter_file_contents),
            dtype=h5py.special_dtype(vlen=str),
        )
        _logger.info(f"Initialized empty file: {output_file}")


def get_data_from_file(
    file_path: str,
    group_name: str,
    dataset_name: str,
    slices: List[slice] = None,
) -> np.ndarray:
    """
    Get the data from the HDF5 file.

    Args:
        file_path: Path to the HDF5 file.
        group_name: Name of the group.
        dataset_name: Name of the dataset.
        slices: List of slices to apply to the dataset. If None, the whole dataset is returned.
    Returns:
        data: np.ndarray
    """
    with h5py.File(file_path, "r") as file:
        dataset = file[f"{group_name}/{dataset_name}"]
        if slices is not None:
            if dataset.ndim != len(slices):
                _logger.error(
                    f"Dataset has {dataset.ndim} dimensions, but {len(slices)} slices were provided."
                )
                raise Exception(
                    f"Dataset has {dataset.ndim} dimensions, but {len(slices)} slices were provided."
                )
            else:
                data = dataset[tuple(slices)]
        else:
            data = dataset[:]

        data = data.astype(np.float64)
    return data


def add_array(
    file_path: str,
    group_name: str,
    dataset_name: str,
    data: np.ndarray,
    attributes: dict = None,
    compression: str = None,
) -> None:
    """
    Adds an array to a Dataset. If the dataset exists, the array is appended.
    Dataset and group is created if not already present.
    Per default no compression is used, bit 'gzip' can be specified.

    Args:
        file_path: Path to the HDF5 file.
        group_name: Name of the group.
        dataset_name: Name of the dataset.
        data: Data to save.
        attributes: Attributes to save.
        compression: Compression to use.
    """
    with h5py.File(file_path, "a") as file:
        if group_name not in file:
            file.create_group(group_name)
            _logger.debug(f"Created group {group_name}")

        group = file[group_name]

        if dataset_name not in group:
            group.create_dataset(
                dataset_name,
                data=data,
                maxshape=(None, *data.shape[1:]),
                chunks=(1, *data.shape[1:]),
                compression=compression,
            )
            dataset = group[dataset_name]
            if attributes:
                for key, value in attributes.items():
                    dataset.attrs[key] = value
            _logger.debug(
                f"Added dataset {dataset_name} to group {group_name} and added data of shape {dataset.shape}"
            )

        else:
            if group[dataset_name].shape[1:] != data.shape[1:]:
                _logger.error(
                    f"Shape of data to add ({data.shape[1:]}) does not match shape of existing dataset ({group[dataset_name].shape[1:]})"
                )
                raise Exception(
                    f"Shape of data to add ({data.shape[1:]}) does not match shape of existing dataset ({group[dataset_name].shape[1:]})"
                )
            if dataset_name == "offset_raw":
                # the data here should not be stacked but averaged
                dataset = group[dataset_name]
                dataset = (dataset + data) / 2
                if attributes:
                    for key, value in attributes.items():
                        dataset.attrs[key] = value
                _logger.debug(
                    f"Appended data to dataset {dataset_name} in group {group_name}. Shape is now {dataset.shape}"
                )

            else:
                dataset = group[dataset_name]
                dataset.resize(dataset.shape[0] + data.shape[0], axis=0)
                dataset[-data.shape[0] :] = data
                if attributes:
                    for key, value in attributes.items():
                        dataset.attrs[key] = value
                _logger.debug(
                    f"Appended data to dataset {dataset_name} in group {group_name}. Shape is now {dataset.shape}"
                )
